/*
 * Harmony.h - effect-mixer for LMMS
 *
 * andrei.v.smirnov@gmail.com
 */


#ifndef _Harmony_H
#define _Harmony_H

//-#define NTONES	sizeof(Harmony::Fractions)/sizeof(Harmony::Fractions[0])

class harmonyTCO;
class HarmonyTrack;

//-typedef QVector<harmonyTCO *> HarmonyTCOVector;

namespace Harmony
{

// Chromatic scale (approximate):
// 1+1/16 -> C#
// 1+1/8 -> D
// 1+1/6 -> D#
// 1+1/5 -> E
// 1+1/3 -> F
// 1+2/5 -> F#
// 1+1/2 -> G
// 1+3/5 -> G#
// 1+2/3 -> A
// 1+7/9 -> A#
// 1+8/9 -> B
	const int Fractions[] = {
1, 8,
1, 6,
3, 16,
1, 5,
2, 9,
1, 4,
3, 10,
5, 16,
1, 3,
3, 8,
2, 5,
5, 12,
4, 9,
1, 2,
8, 15,
9, 16,
3, 5,
5, 8,
2, 3,
3, 4,
4, 5,
5, 6,
8, 9,
15, 16,
1, 1,
16, 15,
9, 8,
6, 5,
5, 4,
4, 3,
3, 2,
8, 5,
5, 3,
16, 9,
15, 8,
2, 1,
9, 4,
12, 5,
5, 2,
8, 3,
3, 1,
16, 5,
10, 3,
4, 1,
9, 2,
5, 1,
16, 3,
6, 1,
8, 1
	};
	const int 
		defaultIndex = 24, // index of the base tone: Fractions[baseIndex]=(1,1)
		nDefaultTones = sizeof(Fractions)/sizeof(Fractions[0])/2; 
	const int defaultHarmonicMods[] = {
1, 2, // octave=-1
3, 5,
5, 8,
2, 3,
3, 4,
4, 5,
5, 6,
8, 9,
1, 1, // octave=0
9, 8,
6, 5,
5, 4,
4, 3,
3, 2,
8, 5,
5, 3,
2, 1 // octve=1 (2**1) 
	};
	const int nDefaultHarmonicMods = sizeof(defaultHarmonicMods)/sizeof(defaultHarmonicMods[0])/2;
	extern QString name;
	const int nMeasures = 4;
	extern int 
		ticksPerMeasure, // = midiTime::ticksPerTact()/nMeasures;
		harmonicFractionsIndex,
		nTacts,  // each tact is subdivided into nMeasures
		nShifts, //=nTacts*nMeasures
		pos, // current position in ticks
		baseIndex, nTones,
//?		frequencyShiftNum, frequencyShiftDen,
		nHarmonicMods,
		*Num, *Den, *harmonicMods;
	extern float 
//-		frequencyShift,//current frequency shift
		*Scale, // Scale[nTones]
		*Shift; // Shift[nTacts*nMeasures]

	void init();
	void cleanup();
	void setScale(const int nTones, const int *fractions);
	void setMods(const int nHarmonicMods, const int *harmonicMods);
//-	void load(char filename[]);

	void saveSettings( QDomDocument & _doc, QDomElement & _parent );
	void loadSettings( const QDomElement & _this );
	void saveState( QDomDocument & _doc, QDomElement & _this );
	void loadState( QDomDocument & _doc, QDomElement & _this );
	void restoreState( const QDomElement & _this );
	float getFrequency( int key, int pos, float modulation );
//-	void addHarmonicScale( HarmonicScale * _hs );
//-	void removeHarmonicScale( HarmonicScale * _hs );

	extern HarmonyTrack	*track;

} ;
#endif
