/*
 * HarmonyTrack.h - class HarmonyTrack, a track which provides harmonic sequence
 *
 * Copyright (c) 2011 Andrei Smirnov <andrei.v.smirnov@gmail.com>
 *
 * This file is part of Linux MultiMedia Studio - http://lmms.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program (see COPYING); if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA.
 *
 */

#ifndef _MODULATION_TRACK_H
#define _MODULATION_TRACK_H

#include <QtGui/QDialog>
#include <QtGui/QPushButton>
#include <QtGui/QRadioButton>
#include <QtGui/QGroupBox>

//-#include "Piano.h"
//-#include "PianoView.h"

#include "track_label_button.h"
//?#include "fade_button.h"

class QLineEdit;
template<class T> class QQueue;
//-class EffectRackView;
class ArpeggiatorView;
class ChordCreatorView;
class EffectRackView;
class knob;
class LcdSpinBox;
class PluginView;
//-class sampleBuffer;
class tabWidget;
class QMdiSubWindow;
class QLabel;

class HarmonyTrackWindow;
class HarmonyTrackView;
class harmonyTCO;

class ChooseHarmonicScaleDialog : public QWidget
{
	Q_OBJECT
	QVector<QRadioButton *> harmonyButtons[Harmony::nMeasures];
	QGroupBox *radioBoxes[Harmony::nMeasures];

	harmonyTCO *m_tco;
	

public slots:
	void update();

public:
	ChooseHarmonicScaleDialog(QWidget * _parent, harmonyTCO * _tco);
	virtual ~ChooseHarmonicScaleDialog();
	void setNmeasures(int n);
	int getNmeasures();
	void setHarmonicFractionIndex(int i, int j);
	int getHarmonicFractionIndex(int i);

signals:
	void currentPatternChanged();
	void closed();

protected:
	virtual void closeEvent( QCloseEvent * _ce );
} ;



class harmonyTCO : public trackContentObject
{
	Q_OBJECT
	mapPropertyFromModel(bool,isRecord,setRecord,m_recordModel);
//-	HarmonicScale *harmonicScale;
	int
		nMeasures, // actual number of measures in TCO
		harmonicFractionIndex[Harmony::nMeasures]; 
//?		frequencyShiftRatio[Harmony::nMeasures][2]; // fractional representation of the frequency shift
//?	float frequencyShift[Harmony::nMeasures]; // = frequencyShiftRatio[0]/frequencyShiftRatio[1]
/* csound/muma/play.py:
		scaleLen=len(scale)
		scaleIndex=keyIndex/scaleLen # octave index
		scaleKey=keyIndex%scaleLen # key index within the scale
		tone=scale[scaleKey]+1. # +1 to simplify scales
		frequency=baseFrequency*frequencyShift*tone*2**scaleIndex
	# where baseFrequency is the lowest frequency: 
	# scales[0].key[0], where scales is a list of scale-s, and
	# scale is a list of fractions defining a scale, for example:
	# (0.,F(1,5),F(1,4),F(1,3),F(2,5),F(1,2),F(3,5),F(2,3),F(3,4),F(4,5))
*/
	
public:
	harmonyTCO( track * _track );
	virtual ~harmonyTCO();

	virtual void changeLength( const MidiTime & _length );
	const QString & harmonyFile() const;

	virtual void saveSettings( QDomDocument & _doc, QDomElement & _parent );
	virtual void loadSettings( const QDomElement & _this );
	inline virtual QString nodeName() const
	{
		return "htco";
	}

//?	inline void setFrequencyShift(int iMeasure, int num, int den) { 
//?		if (num < 0 || den <= 0) {
//?				//+ Display Error 
//?				return;
//?		}
//?		frequencyShiftRatio[iMeasure][0] = num; 
//?		frequencyShiftRatio[iMeasure][1] = den; 
//?		frequencyShift[iMeasure] = (float)num/(float)den;
//?	};
//?	inline float getFrequencyShift(int i) {return frequencyShift[i];};
//?	inline int getFrequencyShiftNum(int i) { return frequencyShiftRatio[i][0]; }
//?	inline int getFrequencyShiftDen(int i) { return frequencyShiftRatio[i][1]; }
	inline void setHarmonicFractionIndex(int iMeasure, int i) { harmonicFractionIndex[iMeasure] = i; }
	inline int getHarmonicFractionIndex(int i) { return harmonicFractionIndex[i]; }
	inline void setNmeasures(int n) { nMeasures = n; }
	inline int getNmeasures() { return nMeasures; }

	MidiTime harmonyLength() const;

	virtual trackContentObjectView * createView( trackView * _tv );

//-signals:
//-	void changed();

public slots:
	void updateLength( bpm_t = 0 );
	void toggleRecord();


private:
	BoolModel m_recordModel;


	friend class harmonyTCOView;

} ;



class harmonyTCOView : public trackContentObjectView
{
	Q_OBJECT
public:
	harmonyTCOView( harmonyTCO * _tco, trackView * _tv );
	virtual ~harmonyTCOView();


public slots:
	void updateHarmony();
	void closeChooseHarmonicScaleDlg();


protected:
	virtual void contextMenuEvent( QContextMenuEvent * _cme );
	virtual void mousePressEvent( QMouseEvent * _me );
	virtual void dragEnterEvent( QDragEnterEvent * _dee );
	virtual void dropEvent( QDropEvent * _de );
	virtual void mouseDoubleClickEvent( QMouseEvent * );
	virtual void paintEvent( QPaintEvent * );


private:
	harmonyTCO * m_tco;
	HarmonyTrackView *m_tv;
	ChooseHarmonicScaleDialog *m_chooseHarmonicScaleDlg;
	QMdiSubWindow * m_subWindow;
	bool m_show;
} ;




class HarmonyTrack : public track
{
	Q_OBJECT
public:
	HarmonyTrack( TrackContainer * _tc );
	virtual ~HarmonyTrack();

	virtual bool play( const MidiTime & _start, const fpp_t _frames,
						const f_cnt_t _frame_base,
							int _tco_num = -1 );
	virtual trackView * createView( TrackContainerView * _tcv );
	virtual trackContentObject * createTCO( const MidiTime & _pos );


	virtual void saveTrackSpecificSettings( QDomDocument & _doc,
							QDomElement & _parent );
	virtual void loadTrackSpecificSettings( const QDomElement & _this );


	virtual QString nodeName() const
	{
		return "harmonytrack";
	}

	void setShift();

private:
	friend class HarmonyTrackView;
	friend class HarmonyTrackWindow;

} ;



class HarmonyTrackView : public trackView
{
	Q_OBJECT
public:
	HarmonyTrackView( HarmonyTrack * _track, TrackContainerView * _tcv );
	virtual ~HarmonyTrackView();

	HarmonyTrackWindow * getHarmonyTrackWindow();
	static HarmonyTrackWindow * topLevelHarmonyTrackWindow();
	void freeHarmonyTrackWindow();

	HarmonyTrack * model()
	{
		return castModel<HarmonyTrack>();
	}


public slots:
	void showEffects();
	void showSetButton();

private slots:
	void setButtonPressed();
//-	void toggleHarmonyWindow( bool _on );

protected:
	void modelChanged();
	virtual QString nodeName() const
	{
		return "HarmonyTrackView";
	}


private:
	HarmonyTrackWindow * m_window;

	static QQueue<HarmonyTrackWindow *> s_windowCache;

	trackLabelButton * m_tlb;
	QPushButton * m_setButton;

	friend class HarmonyTrackWindow;
} ;




class HarmonyTrackWindow : public QWidget, public ModelView,
								public SerializingObjectHook
{
	Q_OBJECT
public:
	HarmonyTrackWindow( HarmonyTrackView * _tv );
	virtual ~HarmonyTrackWindow();

	// parent for all internal tab-widgets
	tabWidget * tabWidgetParent()
	{
		return m_tabWidget;
	}

	HarmonyTrack * model()
	{
		return castModel<HarmonyTrack>();
	}

	const HarmonyTrack * model() const
	{
		return castModel<HarmonyTrack>();
	}

	void setHarmonyTrackView( HarmonyTrackView * _tv );

	static void dragEnterEventGeneric( QDragEnterEvent * _dee );

	virtual void dragEnterEvent( QDragEnterEvent * _dee );
	virtual void dropEvent( QDropEvent * _de );


public slots:
	void textChanged( const QString & _new_name );
	void toggleVisibility( bool _on );
	void updateName();
	void updateHarmonyView();


protected:
	// capture close-events for toggling instrument-track-button
	virtual void closeEvent( QCloseEvent * _ce );
	virtual void focusInEvent( QFocusEvent * _fe );

	virtual void saveSettings( QDomDocument & _doc, QDomElement & _this );
	virtual void loadSettings( const QDomElement & _this );


protected slots:
	void saveSettingsBtnClicked();


private:
	virtual void modelChanged();

	HarmonyTrack * m_track;
	HarmonyTrackView * m_itv;

	// widgets on the top of an instrument-track-window
	tabWidget * m_generalSettingsWidget;
	QLineEdit * m_nameLineEdit;
	knob * m_volumeKnob;
	LcdSpinBox * m_effectChannelNumber;
	QPushButton * m_saveSettingsBtn;


	// tab-widget with all children
	tabWidget * m_tabWidget;
//-	PluginView * m_instrumentView;

	// test-piano at the bottom of every instrument-settings-window

//-	friend class HarmonyView;

} ;

#endif
