#ifndef SINGLE_SOURCE_COMPILE

/*
 * Harmony.cpp - effect mixer for LMMS
 *
 * andrei.v.smirnov@gmail.com
 */

#include <stdio.h>
#include <iostream>//-
using namespace std;//-

#include <math.h>

#include <QtXml/QDomElement>

#include "track.h"
#include "Harmony.h"
#include "Effect.h"
#include "song.h"




namespace Harmony
{
	QString name = "Harmonic Scale";
//-	HarmonicScale *harmonicScale = NULL;// current instrument's scale
//-	HarmonyTCOVector tcos;
	int
		nTacts = 0, 
		nShifts = 0,
		baseIndex = defaultIndex, 
		ticksPerMeasure = MidiTime::ticksPerTact()/nMeasures,
		harmonicFractionsIndex=0,
		nTones = 0,
		nHarmonicMods = 0,
//?		frequencyShiftNum = 0,
//?		frequencyShiftDen = 0,
		pos = -1, // current position in ticks
		*harmonicMods = NULL,
		*Num = NULL,
		*Den = NULL;
	float 
		*Scale = NULL, 
		*Shift = NULL; 
//-		frequencyShift = 1.0; // current frequency shift (harmonic key)
	bool shiftChanged = false;
	HarmonyTrack	*track;
} ;


void Harmony::init() {

	setScale(nDefaultTones, Fractions);
	setMods(nDefaultHarmonicMods, defaultHarmonicMods);
	if (Shift != NULL) delete Shift;
	Shift = NULL;
	nTacts = 0;
	shiftChanged = false;

}

void Harmony::cleanup() {
	delete Num;
	delete Den;
	delete Scale;
	delete Shift;
}

#define SMALL	10.e-20
void Harmony::setScale(const int _nTones, const int *fractions) {
//-	if (_nTones > MAXKEYS) {
//-		cerr << "Scale exceeds max "<<MAXKEYS<<" tones\n";
//-		exit(2);
//-	}
	if (nTones != _nTones) {
		if (Num != NULL) delete Num;
		if (Den != NULL) delete Den;
		if (Scale != NULL) delete Scale;
		nTones = _nTones;
		Num = new int[nTones];
		Den = new int[nTones];
		Scale = new float[sizeof(*Scale)*nTones];
	}
	baseIndex = -1;
	for (int i=0; i<nTones; i++) {
		int j=2*i;
		Num[i] = fractions[j];
		Den[i] = fractions[j+1];
		Scale[i] = (float)Num[i]/(float)Den[i];
		if (fabs(Scale[i]-1.0) < SMALL) 
			baseIndex = i;
	}
cout << "Harmony::setScale:nTones="<<nTones<<", baseIndex="<<baseIndex<<endl;//-
	if (baseIndex < 0) {
		cerr << "Scale ratios should include 1/1 = 1\n";
		exit(2);
	}
}

void Harmony::setMods(const int _nHarmonicMods, const int *_harmonicMods) {
	if (nHarmonicMods != _nHarmonicMods) {
		nHarmonicMods = _nHarmonicMods;
		if (harmonicMods != NULL) delete harmonicMods;
		harmonicMods = new int[2*nHarmonicMods];
	}
	for (int i=0; i<nHarmonicMods; i++) {
		int 
			j=2*i,
			num=_harmonicMods[j],
			den=_harmonicMods[j+1];
		harmonicMods[j] = num;
		harmonicMods[j+1] = den;
		float harmonicFraction = (float)num/(float)den;
		if (fabs(harmonicFraction-1.0) < SMALL)
			harmonicFractionsIndex = i;
	}
}
//#undef SMmods
void Harmony::saveState( QDomDocument & _doc, QDomElement & _this ) { // load old scale
	const int buflen=4096; // string buffer
	QDomElement harmony =_doc.createElement( "scales" );
	_this.appendChild( harmony );
//-	harmony.setAttribute("tacts", nMeasures);
	// Save Harmonic Scale:
	QDomElement instr_scale =_doc.createElement( "instrument" );
	harmony.appendChild( instr_scale );
	instr_scale.setAttribute("size", nTones);
	char data[buflen],*p=data;
	for( int i = 0; i < nTones && p-data < buflen; ++i )
	{
		sprintf(p,"%d %d ",Num[i],Den[i]);
		p += strlen(p);
	}
	QDomText t = _doc.createTextNode(data);
	instr_scale.appendChild(t);
	// Save Harmonic Mods:
	QDomElement harmonic_scale =_doc.createElement( "harmony" );
	harmony.appendChild( harmonic_scale );
	harmonic_scale.setAttribute("size", nHarmonicMods);
	p=data;
	for( int i = 0; i < nHarmonicMods && p-data < buflen; ++i )
	{
		sprintf(p,"%d %d ",harmonicMods[2*i],harmonicMods[2*i+1]);
		p += strlen(p);
	}
	t = _doc.createTextNode(data);
	harmonic_scale.appendChild(t);
}


void Harmony::loadState( QDomDocument & _doc, QDomElement & _this ) { // load old scale

}

void Harmony::restoreState( const QDomElement & _this )
{
	int nMeasures = _this.attribute("tacts").toInt();
	cout << "Harmony::restoreState:tagName="<<_this.tagName().toAscii().constData()<<", nMeasures="<<nMeasures<<endl;//-
	QDomNode node = _this.firstChild();
	while( !node.isNull() ) {
		QDomElement e = node.toElement();
//cout << "\te.tagName="<<e.tagName().toAscii().constData()<<endl;//-
		if (e.tagName() == "instrument") {
			int 
				ntones = e.attribute("size").toInt(),
				*fractions = new int[2*ntones];
			QStringList data = e.text().split(' ');
//cout << "instr_scale:ntones="<<ntones<<", size="<<data.size()<<endl;//-
			for (int i=0; i<ntones; i++) {
				int j=2*i,j1=j+1;
				fractions[j] = data.at(j).toInt();
				fractions[j1] = data.at(j1).toInt();
			}
			setScale( ntones, fractions );
			delete fractions;
		}
		if (e.tagName() == "harmony") {
			int 
				nmods = e.attribute("size").toInt(),
				*hmods = new int[2*nmods];
			QStringList data = e.text().split(' ');
			for (int i=0; i<nmods; i++) {
				int j=2*i,j1=j+1;
				hmods[j] = data.at(j).toInt();
				hmods[j1] = data.at(j1).toInt();
			}
			setMods( nmods, hmods );
			delete hmods;
		}
		node = node.nextSibling();
	}
}


void Harmony::saveSettings( QDomDocument & _doc, QDomElement & _this )
{
}




void Harmony::loadSettings( const QDomElement & _this )
{

}



float Harmony::getFrequency(
	int _key,
	int _pos,
	float instrumentModulation
	) {	

	if (_key >= nTones) return -1.0;

	const float 
		pitch = engine::getSong()->masterPitch(),
		scaleModulation = Harmony::Scale[_key];

	float harmonicModulation = 1.0;

	if (pos >= 0) {
		int ntacts = pos / MidiTime::ticksPerTact();
		if (ntacts < Harmony::nTacts) {
			int
				ipos = pos % MidiTime::ticksPerTact(),
				iMeasure = ipos / Harmony::ticksPerMeasure,
				ishift = ntacts * Harmony::nMeasures + iMeasure;//-
			harmonicModulation = Harmony::Shift[ishift];
		}
	} else if (_pos >=0) {
		int ntacts = _pos / MidiTime::ticksPerTact();
		if (ntacts < Harmony::nTacts) {
			int 
				ipos = _pos % MidiTime::ticksPerTact(),
				iMeasure = ipos / Harmony::ticksPerMeasure,
				ishift = ntacts * Harmony::nMeasures + iMeasure;//-
			harmonicModulation = Harmony::Shift[ishift];
		}
	}

	return BaseFreq * (1. + pitch/12.)  
	* instrumentModulation 
	* harmonicModulation 
	* scaleModulation;
}
//-HarmonicScale::HarmonicScale(QString _name) 
//-{
//-	name = _name;
//-	num = den = NULL;
//-	shift = NULL;
//-}
//-
//-
//-
//-
//-HarmonicScale::~HarmonicScale() 
//-{
//-	ntones = 0;
//-	delete num;
//-	delete den;
//-	delete shift;
//-}
//-
//-
//-
//-void HarmonicScale::setScale(int _ntones, int *_num, int *_den) {
//-	if (num != NULL) delete num;
//-	if (den != NULL) delete den;
//-//-	num = _num; den = _den;
//-	ntones = _ntones + 1;
//-	num = new int[ntones]; 
//-	den = new int[ntones];
//-	num[0] = den[0] = 1;
//-	if (shift != NULL) delete shift;
//-	shift = new float[ntones];
//-	shift[0] = 1.0;
//-	for (int i=1; i<ntones; i++) {
//-		num[i] = _num[i-1];
//-		den[i] = _den[i-1];
//-		shift[i] = (float)num[i]/(float)den[i];
//-	}
//-}

#endif
