/*
 * HarmonyTrack.cpp - implementation of class HarmonyTrack, a track which
 *                    provides harmonic scales other than tempered clavier
 *
 * Copyright (c) 2011 Andrei Smirnov <andrei.v.smirnov@gmail.com>
 *
 * This file is part of Linux MultiMedia Studio - http://lmms.sourceforge.net
 *
 *
 */

#include "stdio.h"
#include <iostream>//-
using namespace std;//-

//-#include <QtCore/QDir>
//-#include <QtCore/QFile>
//-#include <QtGui/QApplication>
//-#include <QtGui/QCloseEvent>
//-#include <QtGui/QMessageBox>


#include <QtCore/QQueue>
#include <QtXml/QDomElement>
#include <QtGui/QDropEvent>
#include <QtGui/QFileDialog>
#include <QtGui/QMenu>
#include <QtGui/QLayout>
#include <QtGui/QLineEdit>
#include <QtGui/QMdiArea>
#include <QtGui/QMdiSubWindow>
#include <QtGui/QPainter>
#include <QtGui/QPushButton>
#include <QtGui/QRadioButton>
#include <QtGui/QLabel>
#include <QtGui/QIntValidator>
#include <QtGui/QScrollArea>
#include <QtGui/QVBoxLayout>
//-#include <QtGui/QScrollBar>

#include "gui_templates.h"
#include "config_mgr.h"
#include "song.h"
#include "embed.h"
#include "engine.h"
#include "FxMixer.h"
#include "FxMixerView.h"
#include "tooltip.h"
#include "AudioPort.h"
#include "PluginView.h"
//-#include "sample_play_handle.h"
//-#include "sample_record_handle.h"
#include "string_pair_drag.h"
#include "knob.h"
#include "LcdSpinBox.h"
#include "led_checkbox.h"
#include "MainWindow.h"
#include "EffectRackView.h"
//-#include "track_label_button.h"
#include "tab_widget.h"
#include "track_label_button.h"
#include "track.h"
#include "Harmony.h"
#include "HarmonyTrack.h"

//-QPushButton *shiftSetButton = NULL; // dirty fix, since I could not
// find the way to communicate between TrackView and TCOView objects

const char * harmony_track_help = QT_TRANSLATE_NOOP( "HarmonyTrack",
						"With this knob you can set "
						"the harmony of the opened "
						"channel.");

const int INSTRUMENT_WIDTH	= 254;
const int INSTRUMENT_HEIGHT	= INSTRUMENT_WIDTH;
const int PIANO_HEIGHT		= 84;
const int INSTRUMENT_WINDOW_CACHE_SIZE = 8;


ChooseHarmonicScaleDialog::ChooseHarmonicScaleDialog(QWidget *_parent, harmonyTCO * _tco) : 
	QWidget(_parent),
	m_tco ( _tco )
//	harmonicScale( NULL ),
{
//?	QScrollArea *scrollArea = new QScrollArea( this );
//?	scrollArea->setVerticalScrollBarPolicy( Qt::ScrollBarAlwaysOn );
//?	scrollArea->setHorizontalScrollBarPolicy( Qt::ScrollBarAlwaysOff );
//?	scrollArea->setPalette( QApplication::palette( scrollArea ) );
//?	scrollArea->setMinimumHeight( 64 );
//?	QWidget * w = new QWidget();
//?	scrollArea->setWidget( w );
	QGridLayout *layout = new QGridLayout;
//-	QScrollBar *scroll = new QScrollBar();
//-//-	scroll->SetOrientation(Qt::Vertical());
//-//-	scroll->SetMinimum(2);
//-//-	scroll->SetMaximum(10);
//-	scroll->setTracking(0);
//-	scroll->resize(40,20);
//-	layout->addWidget(scroll);
//?	scrollArea->addWidget( layout );
	int
		ncols = Harmony::nMeasures, 
		nrows = Harmony::nHarmonicMods,
		nMeasures = _tco->getNmeasures();
	QGroupBox *labelBox = new QGroupBox(tr("Ratios"));
	QVBoxLayout *vbox = new QVBoxLayout();
//?	vbox->setSpacing(14);
	for (int irow = 0; irow < nrows; irow++) {
		char labelText[8];
		int ifrac = nrows - irow -1; // reverse so that low tones are at the bottom
		sprintf(labelText,"%d/%d",Harmony::harmonicMods[2*ifrac],Harmony::harmonicMods[2*ifrac+1]);
		QLabel *label = new QLabel(labelText);
		vbox->addWidget( label );
	}
	labelBox->setLayout(vbox);
	layout->addWidget(labelBox, 0, 0);
	for (int icol = 0; icol < ncols; icol++) {
		char scol[16];
		sprintf(scol,"%d",icol);
		int ihf = nrows - getHarmonicFractionIndex(icol)-1;
		QGroupBox *radioBox = new QGroupBox(tr(scol));
		radioBox->setCheckable(true);
		if (icol >= nMeasures) {
			radioBox->setChecked(false);
		} else {
			radioBox->setChecked(true);
		}
		radioBoxes[icol] = radioBox;
		QVBoxLayout *vbox = new QVBoxLayout();
		vbox->setSpacing(14);
		for (int irow = 0; irow < nrows; irow++) {
			QRadioButton *checkButton = new QRadioButton();
			if (icol < nMeasures && irow == ihf)
				checkButton->setChecked(true);
			else
				checkButton->setChecked(false);
			vbox->addWidget(checkButton); 
			harmonyButtons[icol].append(checkButton);
		}
		radioBox->setLayout(vbox);
		layout->addWidget(radioBox, 0, icol+1);
	}

	QPushButton *OKbutton = new QPushButton( this );
	OKbutton->setText( tr( "OK" ) );
	connect(OKbutton, SIGNAL( clicked() ), this, SLOT( update() ));

	layout->addWidget( OKbutton, nrows, 0 );
	this->setLayout( layout );
	setFocus();
}


ChooseHarmonicScaleDialog::~ChooseHarmonicScaleDialog()
{
	for (int i=0; i<Harmony::nMeasures; i++)
		harmonyButtons[i].clear();
}


void ChooseHarmonicScaleDialog::setNmeasures(int n) { 
	m_tco->setNmeasures(n); 
}


int ChooseHarmonicScaleDialog::getNmeasures() {
	return m_tco->getNmeasures();
}


void ChooseHarmonicScaleDialog::setHarmonicFractionIndex(int i, int j) {
	m_tco->setHarmonicFractionIndex(i,j); 
}


int ChooseHarmonicScaleDialog::getHarmonicFractionIndex(int i) { 
	return m_tco->getHarmonicFractionIndex(i); 
}


void ChooseHarmonicScaleDialog::update()
{
	int nMeasures = 0;
	for (int iMeasure=0; iMeasure<Harmony::nMeasures; iMeasure++) {
		if (radioBoxes[iMeasure]->isChecked() == false) break;
		int ifrac = Harmony::nHarmonicMods-1;
		for (
			QVector<QRadioButton *>::iterator button = harmonyButtons[iMeasure].begin();
			button != harmonyButtons[iMeasure].end();
			button++
		) {
			if((*button)->isChecked()) {
				setHarmonicFractionIndex(iMeasure, ifrac);
				nMeasures++;
			}
			ifrac--;
		}
	}
	setNmeasures(nMeasures);
	emit( currentPatternChanged() );
	QWidget::update();
	close();
}

void ChooseHarmonicScaleDialog::closeEvent( QCloseEvent * _ce )
{
	_ce->ignore();
	emit closed();
	QWidget::update();
}


harmonyTCO::harmonyTCO( track * _track ) :
	trackContentObject( _track )
{
#ifndef SNALL
#define SMALL 1.e-20
#endif
	nMeasures = Harmony::nMeasures;
	for (int i=0; i<Harmony::nMeasures; i++) {
//?		frequencyShiftRatio[i][0] = frequencyShiftRatio[i][1] = 1.;
//?		frequencyShift[i] = (float)frequencyShiftRatio[i][0]/(float)frequencyShiftRatio[i][1];
		harmonicFractionIndex[i] = Harmony::harmonicFractionsIndex;
	}
}




harmonyTCO::~harmonyTCO()
{
}




void harmonyTCO::changeLength( const MidiTime & _length )
{
	trackContentObject::changeLength( qMax( static_cast<int32_t>( _length ),
							DefaultTicksPerTact ) );
}




const QString & harmonyTCO::harmonyFile() const
{
	static const QString filename = "tmp";//DDD
//-	return m_sampleBuffer->audioFile();
	return filename;//DDD
}




void harmonyTCO::toggleRecord()
{
	m_recordModel.setValue( !m_recordModel.value() );
	emit dataChanged();
}




void harmonyTCO::updateLength( bpm_t )
{
	changeLength( harmonyLength() );
}




MidiTime harmonyTCO::harmonyLength() const
{
//-	return (int)( m_sampleBuffer->frames() / engine::framesPerTick() );
	return 0; ///DDD
}




void harmonyTCO::saveSettings( QDomDocument & _doc, QDomElement & _this )
{
	if( _this.parentNode().nodeName() == "clipboard" )
	{
		_this.setAttribute( "pos", -1 );
	}
	else
	{
		_this.setAttribute( "pos", startPosition() );
	}
	_this.setAttribute( "len", length() );
	_this.setAttribute( "muted", isMuted() );
	_this.setAttribute( "src", harmonyFile() );
	_this.setAttribute( "tacts", nMeasures );

	char data[256],*p=data;
	for (int i=0; i<nMeasures; i++) {
		sprintf(p,"%d ",harmonicFractionIndex[i]);
		p += strlen(p);
	}
	QDomText t = _doc.createTextNode(data);
	_this.appendChild(t);

	if( harmonyFile() == "" )
	{
	}
	// TODO: start- and end-frame
}




void harmonyTCO::loadSettings( const QDomElement & _this )
{
	if( _this.attribute( "pos" ).toInt() >= 0 )
	{
		movePosition( _this.attribute( "pos" ).toInt() );
	}
	changeLength( _this.attribute( "len" ).toInt() );
	setMuted( _this.attribute( "muted" ).toInt() );
	nMeasures = _this.attribute("tacts").toInt();
	QStringList data = _this.text().split(' ');
	for (int i=0; i<nMeasures; i++) {
		int j = data.at(i).toInt();
//?			num = Harmony::harmonicMods[2*j],
//?			den = Harmony::harmonicMods[2*j+1];
		harmonicFractionIndex[i] = j;
//?		setFrequencyShift(i, num, den);
	}
//-	emit(changed());
}




trackContentObjectView * harmonyTCO::createView( trackView * _tv )
{
	return new harmonyTCOView( this, _tv );
}



harmonyTCOView::harmonyTCOView( harmonyTCO * _tco, trackView * _tv ) :
	trackContentObjectView( _tco, _tv ),
	m_tco( _tco ),
	m_tv( (HarmonyTrackView *) _tv ),
	m_chooseHarmonicScaleDlg( NULL ),
	m_show( true )
{
//?	connect( m_tco, SIGNAL( harmonyChanged() ), this, SLOT( updateHarmony() ) );

	m_chooseHarmonicScaleDlg = new ChooseHarmonicScaleDialog(this, m_tco);
	m_subWindow = engine::mainWindow()->workspace()->addSubWindow( 
                m_chooseHarmonicScaleDlg );
	
	Qt::WindowFlags flags = m_subWindow->windowFlags();
	flags |= Qt::MSWindowsFixedSizeDialogHint;
	flags &= ~Qt::WindowMaximizeButtonHint;
	m_subWindow->setWindowFlags( flags );
	m_subWindow->setFixedSize( m_subWindow->size() );

//?	m_subWindow->setWindowIcon( m_harmonicScaleDlg->windowIcon() );

	connect( m_chooseHarmonicScaleDlg, SIGNAL( closed() ),
		this, SLOT( closeChooseHarmonicScaleDlg() ) );

// 
//...
// In updateHarmonicScale: transfer info on itact and m_chooseHarmonicScale to tact:scale:offset array
	connect( m_chooseHarmonicScaleDlg, SIGNAL( currentPatternChanged() ),
		this, SLOT( updateHarmony() ) );

//-	connect( m_tco, SIGNAL(changed()), m_tv, SLOT(showSetButton())); 

//-	if (m_tco->getHarmonicScale() != NULL)
//-		m_chooseHarmonicScaleDlg->setHarmonicScale(m_tco->getHarmonicScale());

	m_subWindow->hide();
}


//-void harmonyTCOView::toggleSubWindow() { }
void harmonyTCOView::closeChooseHarmonicScaleDlg()
{
	m_subWindow->hide();
	m_show = true;
	update();
}



harmonyTCOView::~harmonyTCOView()
{
	delete m_chooseHarmonicScaleDlg;
}




void harmonyTCOView::updateHarmony()
{
	int nMeasures = m_chooseHarmonicScaleDlg->getNmeasures();
	m_tco->setNmeasures(nMeasures);
	
	HarmonyTrack *htrack = (HarmonyTrack*) m_tco->getTrack();
	htrack->setShift();
	m_tv->showSetButton();
}





void harmonyTCOView::contextMenuEvent( QContextMenuEvent * _cme )
{
	QMenu contextMenu( this );
	if( fixedTCOs() == false )
	{
		contextMenu.addAction( embed::getIconPixmap( "cancel" ),
					tr( "Delete (middle mousebutton)" ),
						this, SLOT( remove() ) );
		contextMenu.addSeparator();
		contextMenu.addAction( embed::getIconPixmap( "edit_cut" ),
					tr( "Cut" ), this, SLOT( cut() ) );
	}
	contextMenu.addAction( embed::getIconPixmap( "edit_copy" ),
					tr( "Copy" ), m_tco, SLOT( copy() ) );
	contextMenu.addAction( embed::getIconPixmap( "edit_paste" ),
					tr( "Paste" ), m_tco, SLOT( paste() ) );
	contextMenu.addSeparator();
	contextMenu.addAction( embed::getIconPixmap( "muted" ),
				tr( "Mute/unmute (<Ctrl> + middle click)" ),
						m_tco, SLOT( toggleMute() ) );
	contextMenu.addAction( embed::getIconPixmap( "record" ),
				tr( "Set/clear record" ),
						m_tco, SLOT( toggleRecord() ) );
	constructContextMenu( &contextMenu );

	contextMenu.exec( QCursor::pos() );
}




void harmonyTCOView::dragEnterEvent( QDragEnterEvent * _dee )
{
	if( stringPairDrag::processDragEnterEvent( _dee,
					"harmonyfile,harmonydata" ) == false )
	{
		trackContentObjectView::dragEnterEvent( _dee );
	}
}




void harmonyTCOView::dropEvent( QDropEvent * _de )
{
//-	if( stringPairDrag::decodeKey( _de ) == "harmonyfile" )
//-	{
//-		m_tco->setHarmonyFile( stringPairDrag::decodeValue( _de ) );
//-		_de->accept();
//-	}
//-	else if( stringPairDrag::decodeKey( _de ) == "harmonydata" )
//-	{
//-//-		m_tco->m_sampleBuffer->loadFromBase64(
//-//-					stringPairDrag::decodeValue( _de ) );
//-		m_tco->updateLength();
//-		update();
//-		_de->accept();
//-		engine::getSong()->setModified();
//-	}
//-	else
//-	{
		trackContentObjectView::dropEvent( _de );
//-	}
}




void harmonyTCOView::mousePressEvent( QMouseEvent * _me )
{
	if( _me->button() == Qt::LeftButton &&
		_me->modifiers() & Qt::ControlModifier &&
		_me->modifiers() & Qt::ShiftModifier )
	{
		m_tco->toggleRecord();
	}
	else
	{
		trackContentObjectView::mousePressEvent( _me );
	}
}





void harmonyTCOView::mouseDoubleClickEvent( QMouseEvent * )
{
	if( m_show )
	{
		m_subWindow->show();
		m_subWindow->raise();
		m_show = false;
	}
	else
	{
		m_subWindow->hide();
		m_show = true;
	}
}




void harmonyTCOView::paintEvent( QPaintEvent * _pe )
{
	QPainter p( this );

	QLinearGradient grad( 0, 0, 0, height() );
	if( isSelected() )
	{
		grad.setColorAt( 1, QColor( 0, 0, 224 ) );
		grad.setColorAt( 0, QColor( 0, 0, 128 ) );
	}
	else
	{
		grad.setColorAt( 0, QColor( 96, 96, 96 ) );
		grad.setColorAt( 1, QColor( 16, 16, 16 ) );
	}
	p.fillRect( _pe->rect(), grad );

	p.setPen( QColor( 0, 0, 0 ) );
//-	p.setPen( QColor( 255, 0, 0 ) );//DDD
	p.drawRect( 0, 0, width()-1, height()-1 );
//-	p.setFont( pointSize<6>( p.font() ) );//DDD
//-	p.drawText( 1, p.fontMetrics().height() - 1, "HARMONY" );//DDD
	if( m_tco->getTrack()->isMuted() || m_tco->isMuted() )
	{
		p.setPen( QColor( 128, 128, 128 ) );
	}
	else
	{
		p.setPen( QColor( 64, 224, 160 ) );
	}
	QRect r = QRect( 1, 1,
			qMax( static_cast<int>( m_tco->harmonyLength() *
				pixelsPerTact() / DefaultTicksPerTact ), 1 ),
								height() - 4 );
	p.setClipRect( QRect( 1, 1, width() - 2, height() - 2 ) );
//-	m_tco->m_sampleBuffer->visualize( p, r, _pe->rect() );
	if( r.width() < width() - 1 )
	{	int 
			nMeasures = m_tco->getNmeasures(),
			totWidth = width() - 2 - r.x() + r.width(),
			dx = (int)((float)totWidth/(float)nMeasures),
			x0 = r.x() + r.width(), 
			x = x0 + dx,
			dy = r.height(), //2,
			y0 = r.y() + dy +4, // + 8,
			y = y0;
		
		for (int i=0; i<nMeasures; i++) {
			int ihf = m_tco->getHarmonicFractionIndex(i);
			float scale = Harmony::Scale[ihf];
			y = y0-dy*scale;
			p.drawLine( x0, y, x, y );
			x0 = x;
			x = x0 + dx;
		}
//-		p.drawLine( r.x() + r.width(), r.y() + r.height() / 2,
//-				width() - 2, r.y() + r.height() / 2 );
	}

	p.translate( 0, 0 );
	if( m_tco->isMuted() )
	{
		p.drawPixmap( 3, 8, embed::getIconPixmap( "muted", 16, 16 ) );
	}
	if( m_tco->isRecord() )
	{
		p.setFont( pointSize<6>( p.font() ) );
		p.setPen( QColor( 224, 0, 0 ) );
		p.drawText( 9, p.fontMetrics().height() - 1, "Rec" );
		p.setBrush( QBrush( QColor( 224, 0, 0 ) ) );
		p.drawEllipse( 4, 5, 4, 4 );
	}
}






HarmonyTrack::HarmonyTrack( TrackContainer * _tc ) :
	track( track::HarmonyTrack, _tc )
//-	m_volumeModel( DefaultVolume, MinVolume, MaxVolume, 1.0, this,
//-							tr( "Volume" ) )
{
	setName( tr( "Harmony" ) );
//-	for( int itco = 0; itco < numOfTCOs(); ++itco )
//-	{ // First pass to find array size
//-		trackContentObject * tco = getTCO( itco );
//-		harmonyTCO * tco = (harmonyTCO *) getTCO( itco );
//-		int 
//-			ticksPerTact = (int)MidiTime::ticksPerTact(),
//-			itact = tco->startPosition()/ticksPerTact,
//-			length = (int)tco->length(),
//-			ntacts = length/ticksPerTact;
//-			//Find the respective TCO in Harmony::track buffer
//-	}
	Harmony::track = this;
}




HarmonyTrack::~HarmonyTrack()
{
	engine::mixer()->removePlayHandles( this );
//-	Harmony::cleanup();
}




bool HarmonyTrack::play( const MidiTime & _start, const fpp_t _frames,
						const f_cnt_t _offset,
							int /*_tco_num*/ )
{
	bool played_a_note = false;
	for( int i = 0; i < numOfTCOs(); ++i )
	{
		trackContentObject * tco = getTCO( i );
		if( tco->startPosition() != _start )
		{
			continue;
		}
	}

	return played_a_note;
}




trackView * HarmonyTrack::createView( TrackContainerView * _tcv )
{
	return new HarmonyTrackView( this, _tcv );
}




trackContentObject * HarmonyTrack::createTCO( const MidiTime & _pos )
{
	
	harmonyTCO *tco = new harmonyTCO( this );
//-	int 
//-		ticksPerTact = (int)MidiTime::ticksPerTact(),
//-		itact = tco->startPosition()/ticksPerTact,
//-		length = (int)tco->length(),
//-		ntacts = length/ticksPerTact;
	return tco;
}




void HarmonyTrack::saveTrackSpecificSettings( QDomDocument & _doc,
							QDomElement & _this )
{
//?	int nTCOs = numOfTCOs();
//?	_this.setAttribute( "size", nTCOs );
//?	Harmony::nTCOs = nTCOs;
//?#if 0
//?	_this.setAttribute( "icon", tlb->pixmapFile() );
//?#endif
//?//-	m_volumeModel.saveSettings( _doc, _this, "vol" );
//?	
//?	for( int itco = 0; itco < nTCOs; ++itco )
//?	{ // First pass to find array size
//?//?		trackContentObject * tco = getTCO( itco );
//?		harmonyTCO * tco = (harmonyTCO *) getTCO( itco );
//?		int 
//?			nmeasures = tco->getNmeasures(), // actual measures in TCO
//?			ticksPerTact = (int)MidiTime::ticksPerTact(),
//?			itact = tco->startPosition()/ticksPerTact,
//?			length = (int)tco->length(),
//?			ntacts = length/ticksPerTact;
//?		char data[256],*p;
//?		QDomElement e = _doc.createElement( "htco" );
//?		e.setAttribute("tact", itact );
//?		e.setAttribute("tacts", ntacts );
//?		e.setAttribute("tacts", nmeasures );
//?		p=data;
//?		for (int i=0; i<nmeasures; i++) {
//?			int 
//?				num = tco->getFrequencyShiftNum(i),
//?				den = tco->getFrequencyShiftDen(i);
//?			sprintf(p,"%d %d ",num,den);
//?			p += strlen(p);
//?		}
//?		QDomText t = _doc.createTextNode(data);
//?	   e.appendChild(t);
//?//?		e.setAttribute( "name", m_instrument->descriptor()->name );
//?//?		SerializingObject::saveState( _doc, e );
//?		_this.appendChild( e );
//?	}
}




void HarmonyTrack::loadTrackSpecificSettings( const QDomElement & _this )
{
//-	int nTCOs = _this.attribute("size").toInt();
//?	Harmony::initTrack(nTCOs);
//?	QDomNode node = _this.firstChild();
//?	int itco = 0;
//?	while( !node.isNull() )
//?	{
//?		if (node.hasAttributes()) {
//?//			int nmeasures = node.attribute("tacts").toInt();
//?		}
//?		if( node.isElement() )
//?		{
//?			QDomElement e = node.toElement();
//?	if (e.hasAttribute("tacts") && 
//?		e.hasAttribute("tact") &&
//?		e.hasAttribute("tacts")
//?	) {	int 
//?			nMeasures = e.attribute("tacts").toInt(),
//?			itact = e.attribute("tact").toInt(),
//?			ntacts= e.attribute("tacts").toInt();
//?<<", itact="<<itact //-
//?<<", ntacts="<<ntacts //-
//?<<endl;//-
//?		Harmony::setTCO(itco, nMeasures, itact, ntacts);
//?	
//?		itco++;
//?
//?
//?//?		tco->setNmeasures(nMeasures);
//?	}
//?		}
//?		node = node.nextSibling();
//?	}
//?
//-	m_volumeModel.loadSettings( _this, "vol" );
}




void HarmonyTrack::setShift() 
{
	int
		nTCOs = numOfTCOs(), 
		nTacts = 0;
	for( int itco = 0; itco < nTCOs; ++itco )
	{ // First pass to find array size
		harmonyTCO * tco = (harmonyTCO *) getTCO( itco );
		int 
			ticksPerTact = (int)MidiTime::ticksPerTact(),
			itact = tco->startPosition()/ticksPerTact,
			length = (int)tco->length(),
			ntacts = length/ticksPerTact;
		if (itact + ntacts > nTacts) nTacts = itact + ntacts;
	}
	Harmony::nTacts = nTacts;
	int nShifts = nTacts * Harmony::nMeasures;
	if (nShifts != Harmony::nShifts) {
		if (Harmony::Shift != NULL) delete Harmony::Shift;
		Harmony::nShifts = nShifts;
		Harmony::Shift = new float[nShifts];
	}
//-cout << "HarmonyTrack::setShift:nTCOs="<<nTCOs<<", nTacts="<<nTacts<<", nShifts="<<Harmony::nShifts<<endl;cout.flush();//-
	for( int itco = 0; itco < nTCOs; ++itco )
	{ // Second pass to fill out the array
		harmonyTCO * tco = (harmonyTCO *) getTCO( itco );
		int 
			nMeasures = tco->getNmeasures(), // actual measures in TCO
			ticksPerTact = (int)MidiTime::ticksPerTact(),
			itact = tco->startPosition()/ticksPerTact,
			iShiftStart = itact*Harmony::nMeasures,
			iShift = iShiftStart,
			length = (int)tco->length(),
			ntacts = length/ticksPerTact;
			
//cout << "Harmony::setShift:itco="<<itco<<", tco->getMeasures="<<tco->getNmeasures()<<endl;//-
		for (int i=0; i<Harmony::nMeasures; i++) {
			int ipos = (i*nMeasures)/Harmony::nMeasures,
				ihf2 = 2*tco->getHarmonicFractionIndex(ipos);
			float frequencyShift = (float)Harmony::harmonicMods[ihf2]
			                     / (float)Harmony::harmonicMods[ihf2+1];
			for (int j=0; j<ntacts; j++) {
				Harmony::Shift[iShift] = frequencyShift;//tco->getFrequencyShift();

//cout << "Harmony::setShift:j="<<j<<", Harmony::Shift[iShift="<<iShift<<"]="<<frequencyShift<<endl;//-
				iShift++;
			}
		}
	}
//?	shiftSetButton->hide();
}



// #### ITV:


QQueue<HarmonyTrackWindow *> HarmonyTrackView::s_windowCache;



HarmonyTrackView::HarmonyTrackView( HarmonyTrack * _t, TrackContainerView * _tcv ) :
	trackView( _t, _tcv )
{
	setFixedHeight( 32 );

	m_tlb = new trackLabelButton( this, getTrackSettingsWidget() );
	m_tlb->setCheckable( true );
//-	connect( m_tlb, SIGNAL( clicked( bool ) ), this, SLOT( showEffects() ) );
	m_tlb->setIcon( embed::getIconPixmap( "harmony_track" ) );
	m_tlb->move( 3, 1 );
	m_tlb->show();
//?	connect( m_tlb, SIGNAL( toggled( bool ) ), this, SLOT( toggleInstrumentWindow( bool ) ) );
	connect( _t, SIGNAL( nameChanged() ), m_tlb, SLOT( update() ) );

	m_setButton = new QPushButton(tr("Set"), this);
	m_setButton->setGeometry(
			DEFAULT_SETTINGS_WIDGET_WIDTH+20, 2, 8+40, 28 );
	m_setButton->show();
	connect( m_setButton, SIGNAL( pressed() ),
				this, SLOT( setButtonPressed() ) );

	setModel( _t );
}




HarmonyTrackView::~HarmonyTrackView()
{
//-	m_effWindow->deleteLater();
}


void HarmonyTrackView::showSetButton() {
	m_setButton->show();
}


void HarmonyTrackView::setButtonPressed()
{
	model()->setShift();
//?	m_setButton->hide();
}



void HarmonyTrackView::showEffects()
{
//-	if( m_effWindow->isHidden() )
//-	{
//-		m_effectRack->show();
//-		m_effWindow->show();
//-		m_effWindow->raise();
//-	}
//-	else
//-	{
//-		m_effWindow->hide();
//-	}
}



void HarmonyTrackView::modelChanged()
{
	trackView::modelChanged();
}



class fxLineLcdSpinBox : public LcdSpinBox 
{
	public:
		fxLineLcdSpinBox( int _num_digits, QWidget * _parent, 
				const QString & _name ) :
			LcdSpinBox( _num_digits, _parent, _name ) {}

	protected:
		virtual void mouseDoubleClickEvent ( QMouseEvent * _me )
		{
			engine::fxMixerView()->setCurrentFxLine( model()->value() );
			//engine::getFxMixerView()->raise();
		}
};




// #### ITW:
HarmonyTrackWindow::HarmonyTrackWindow( HarmonyTrackView * _itv ) :
	QWidget(),
	ModelView( NULL, this ),
	m_track( _itv->model() ),
	m_itv( _itv )
//-	m_instrumentView( NULL )
{
	setAcceptDrops( true );

	// init own layout + widgets
	setFocusPolicy( Qt::StrongFocus );
	QVBoxLayout * vlayout = new QVBoxLayout( this );
	vlayout->setMargin( 0 );
	vlayout->setSpacing( 0 );

	m_generalSettingsWidget = new tabWidget( tr( "GENERAL SETTINGS" ),
									this );
	m_generalSettingsWidget->setFixedHeight( 90 );

	// setup line-edit for changing channel-name
	m_nameLineEdit = new QLineEdit( m_generalSettingsWidget );
	m_nameLineEdit->setFont( pointSize<8>(
						m_nameLineEdit->font() ) );
	m_nameLineEdit->setGeometry( 10, 16, 230, 18 );
	connect( m_nameLineEdit, SIGNAL( textChanged( const QString & ) ),
				this, SLOT( textChanged( const QString & ) ) );


	// setup volume-knob
	m_volumeKnob = new knob( knobBright_26, m_generalSettingsWidget,
						tr( "Harmony volume" ) );
	m_volumeKnob->setVolumeKnob( true );
	m_volumeKnob->move( 10, 44 );
	m_volumeKnob->setHintText( tr( "Volume:" ) + " ", "%" );
	m_volumeKnob->setLabel( tr( "VOL" ) );

	m_volumeKnob->setWhatsThis( tr( harmony_track_help ) );



	// setup spinbox for selecting FX-channel
	m_effectChannelNumber = new fxLineLcdSpinBox( 2, m_generalSettingsWidget,
						tr( "FX channel" ) );
	m_effectChannelNumber->setLabel( tr( "FX CHNL" ) );

	m_saveSettingsBtn = new QPushButton( embed::getIconPixmap(
							"project_save" ), "",
						m_generalSettingsWidget );
	m_saveSettingsBtn->setGeometry( m_effectChannelNumber->x() +
					m_effectChannelNumber->width() + 20, 44,
					32, 32 );
	connect( m_saveSettingsBtn, SIGNAL( clicked() ), this,
					SLOT( saveSettingsBtnClicked() ) );
	toolTip::add( m_saveSettingsBtn, 
		tr( "Save current channel settings in a preset-file" ) );
	m_saveSettingsBtn->setWhatsThis(
		tr( "Click here, if you want to save current channel settings "
			"in a preset-file. Later you can load this preset by "
			"double-clicking it in the preset-browser." ) );


	m_tabWidget = new tabWidget( "", this );
	m_tabWidget->setFixedHeight( INSTRUMENT_HEIGHT + 10 );


	// create tab-widgets
	QWidget * instrument_functions = new QWidget( m_tabWidget );
	m_tabWidget->addTab( instrument_functions, tr( "FUNC" ), 2 );

	// setup piano-widget

	vlayout->addWidget( m_generalSettingsWidget );
	vlayout->addWidget( m_tabWidget );

	setModel( _itv->model() );

	updateHarmonyView();

	setFixedWidth( INSTRUMENT_WIDTH );
	resize( sizeHint() );

	QMdiSubWindow * subWin = 
			engine::mainWindow()->workspace()->addSubWindow( this );
	Qt::WindowFlags flags = subWin->windowFlags();
	flags |= Qt::MSWindowsFixedSizeDialogHint;
	flags &= ~Qt::WindowMaximizeButtonHint;
	subWin->setWindowFlags( flags );
	subWin->setWindowIcon( embed::getIconPixmap( "instrument_track" ) );
	subWin->setFixedSize( subWin->size() );
	subWin->hide();
}




HarmonyTrackWindow::~HarmonyTrackWindow()
{
	HarmonyTrackView::s_windowCache.removeAll( this );

//-	delete m_instrumentView;

	if( engine::mainWindow()->workspace() )
	{
		parentWidget()->hide();
		parentWidget()->deleteLater();
	}
}




void HarmonyTrackWindow::setHarmonyTrackView( HarmonyTrackView * _tv )
{
	if( m_itv && _tv )
	{
		m_itv->m_tlb->setChecked( false );
	}
	m_itv = _tv;
}




void HarmonyTrackWindow::modelChanged()
{
	m_track = castModel<HarmonyTrack>();

	m_nameLineEdit->setText( m_track->name() );

	m_track->disconnect( SIGNAL( nameChanged() ), this );
	m_track->disconnect( SIGNAL( instrumentChanged() ), this );

	connect( m_track, SIGNAL( nameChanged() ),
			this, SLOT( updateName() ) );
	connect( m_track, SIGNAL( instrumentChanged() ),
			this, SLOT( updateHarmonyView() ) );
	
//-	m_volumeKnob->setModel( &m_track->m_volumeModel );

	updateName();
}




void HarmonyTrackWindow::saveSettingsBtnClicked()
{
	QFileDialog sfd( this, tr( "Save preset" ), "",
				tr( "XML preset file (*.xpf)" ) );
// NOT NEEDED
}





void HarmonyTrackWindow::updateName()
{
	setWindowTitle( m_track->name() );

	if( m_nameLineEdit->text() != m_track->name() )
	{
		m_nameLineEdit->setText( m_track->name() );
	}
}





void HarmonyTrackWindow::updateHarmonyView()
{
//-	delete m_instrumentView;
//-	if( m_track->m_instrument != NULL )
//-	{
//-		m_instrumentView = m_track->m_instrument->createView(
//-								m_tabWidget );
//-		m_tabWidget->addTab( m_instrumentView, tr( "PLUGIN" ), 0 );
//-		m_tabWidget->setActiveTab( 0 );
//-
//-		modelChanged(); 		// Get the instrument window to refresh
//-		m_track->dataChanged(); // Get the text on the trackButton to change
//-	}
}




void HarmonyTrackWindow::textChanged( const QString & _new_name )
{
	m_track->setName( _new_name );
	engine::getSong()->setModified();
}




void HarmonyTrackWindow::toggleVisibility( bool _on )
{

	if( _on )
	{
		show();
		parentWidget()->show();
		parentWidget()->raise();
	}
	else
	{
		parentWidget()->hide();
	}
}




void HarmonyTrackWindow::closeEvent( QCloseEvent * _ce )
{
	_ce->ignore();
	if( engine::mainWindow()->workspace() )
	{
		parentWidget()->hide();
	}
	else
	{
		hide();
	}
	m_itv->m_tlb->setFocus();
	m_itv->m_tlb->setChecked( false );
}




void HarmonyTrackWindow::focusInEvent( QFocusEvent * )
{
}




void HarmonyTrackWindow::dragEnterEventGeneric( QDragEnterEvent * _dee )
{
	stringPairDrag::processDragEnterEvent( _dee, "instrument,presetfile,"
							"pluginpresetfile" );
}




void HarmonyTrackWindow::dragEnterEvent( QDragEnterEvent * _dee )
{
	dragEnterEventGeneric( _dee );
}




void HarmonyTrackWindow::dropEvent( QDropEvent * _de )
{
	QString type = stringPairDrag::decodeKey( _de );
	QString value = stringPairDrag::decodeValue( _de );
	if( type == "instrument" )
	{
//-		m_track->loadHarmony( value );
		engine::getSong()->setModified();
		_de->accept();
	}
	else if( type == "presetfile" )
	{
//-		multimediaProject mmp( value );
//-		HarmonyTrack::removeMidiPortNode( mmp );
//-		m_track->setSimpleSerializing();
//-		m_track->loadSettings( mmp.content().toElement() );
		engine::getSong()->setModified();
		_de->accept();
	}
	else if( type == "pluginpresetfile" )
	{
//-		const QString ext = fileItem::extension( value );
//-		Harmony * i = m_track->instrument();
//-		if( !i->descriptor()->supportsFileType( ext ) )
//-		{
//-			i = m_track->loadHarmony(
//-					engine::pluginFileHandling()[ext] );
//-		}
//-		i->loadFile( value );
//-		_de->accept();
	}
}




void HarmonyTrackWindow::saveSettings( QDomDocument & _doc,
							QDomElement & _this )
{
	_this.setAttribute( "tab", m_tabWidget->activeTab() );
	MainWindow::saveWidgetState( this, _this );
}




void HarmonyTrackWindow::loadSettings( const QDomElement & _this )
{
	m_tabWidget->setActiveTab( _this.attribute( "tab" ).toInt() );
	MainWindow::restoreWidgetState( this, _this );
	if( isVisible() )
	{
		m_itv->m_tlb->setChecked( true );
	}
}




#include "moc_HarmonyTrack.cxx"

